import requests
import yaml
import psycopg2
import pandas as pd
from pandas import json_normalize
from datetime import datetime
from Funcoes.consultaTabelas import tabela_existe
from Funcoes.criaTabelas import tabela_nova
from Config.pasta_local import pasta_local

# Coleta data e hora para inserir no banco
data_hora = datetime.now()

# Abre o arquivo YAML com as configurações necessários. Validar o caminho que está salvo os arquivos para execução.
nome_arquivo_yaml = pasta_local+'\Config\config.yaml'

if __name__ == '__main__':
    with open(nome_arquivo_yaml,encoding='utf-8') as file:
        config = yaml.safe_load(file)

# Coleta os dados do arquivo YAML
token = config['token']
db_name = config['connector-oltp']['db_name']
db_user = config['connector-oltp']['db_user']
db_password = config['connector-oltp']['db_password']
db_host = config['connector-oltp']['db_host']
db_schema = config['db_schema']

# Informa a tabela que trabalharemos nesta etapa do ETL
tabela = 'clickup_spaces'

# Comando SQL para criar a tabela
query = f'''
        CREATE TABLE {db_schema}.{tabela} (
            space_id VARCHAR(20), 
            space_nome VARCHAR(50), 
            space_privado VARCHAR(20), 
            space_arquivado VARCHAR(20),
            status_id VARCHAR(40), 
            status_descricao VARCHAR(50),
            status_tipo VARCHAR(20), 
            status_index VARCHAR(5), 
            data_hora TIMESTAMP
        );
        '''

# Verifica se já existe a tabela, caso não exista ela é criada
if tabela_existe(tabela,db_name,db_user,db_password,db_host):
      pass
else :
      tabela_nova(query,db_name,db_user,db_password,db_host)

# conecta no banco
try:
    connection = psycopg2.connect(
        user= db_user,
        password= db_password,
        host= db_host,
        port= "5432",
        database= db_name
    )
    cursor = connection.cursor()

    # Testando a conexão
    cursor.execute("SELECT version();")
    record = cursor.fetchone()

except Exception as error:
    pass

# Criando um novo cursor
cursor = connection.cursor()

# Consulta para selecionar os dados inseridos (ajuste conforme necessário)
select_query = f'SELECT DISTINCT time_id FROM {db_schema}.clickup_team;'

# Executando a consulta SELECT
cursor.execute(select_query)

# Obtendo e imprimindo os resultados
results = cursor.fetchall()

# Iterando sobre a resultado e criando uma lista
lista_times = [row[0] for row in results]

# função para invocar a consutla das workspaces dos times, trabalhará com loop caso tenha mais de um time
def requisicao(token_acesso, team_id):
    url = f'https://api.clickup.com/api/v2/team/{team_id}/space'
    # estrutura os headers conforme documentação utilizando o token salvo na pasta Config - arquivo config.yaml
    headers = {'Authorization': f'{token_acesso}'}
    # executa o request na url informando os headers
    response = requests.get(url, headers=headers)
    # coleta dos dados da respostsa
    data = response.json()
    spaces = data['spaces']
    df = json_normalize(spaces)
    return df

for time in lista_times:
    df = requisicao(token,time)
    

# expandindo coluna statuses
df_status_expandido = df.explode('statuses')

# aplicando o json para tabela
df_expandido = df_status_expandido['statuses'].apply(pd.Series)

# renomeia colunas do status
df_expandido.rename(columns={'id': 'status_id', 'status' : 'status_descricao', 'type' : 'status_tipo', 'orderindex' : 'status_index'}, inplace=True)

# concatena por linha (join) entre a tabela inicial e colunas expandidas do status
df_final = pd.concat([df.drop(columns='statuses'), df_expandido], axis=1)

# renomeia as colunas da tabela final
df_final.rename(columns={'id': 'space_id', 'name' : 'space_nome', 'private' : 'space_privado', 'archived' : 'space_arquivado'}, inplace=True)

# adiciona a coluna data_hora na tabela final
df_final['data_hora'] = data_hora

# seleção das colunas do df para carregamento no banco
df_carregamento = df_final[['space_id', 'space_nome', 'space_privado', 'space_arquivado','status_id', 'status_descricao','status_tipo', 'status_index', 'data_hora']]

# Fechando o cursor atual
cursor.close()
# Criando um novo cursor
cursor = connection.cursor()

# Executando a consulta TRUNCATE TABLE
truncate_query = f'TRUNCATE TABLE {db_schema}.{tabela};'
cursor.execute(truncate_query)

# Confirmando as alterações
connection.commit()

# Fechando o cursor atual
cursor.close()
# Criando um novo cursor
cursor = connection.cursor()

# Preparando a consulta SQL para inserção
insert_query = f'INSERT INTO {db_schema}.{tabela}  (space_id, space_nome, space_privado, space_arquivado, status_id, status_descricao, status_tipo, status_index, data_hora) ' + 'VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s);'

# Iterando sobre as linhas do DataFrame e inserindo no banco
for index, row in df_carregamento.iterrows():
    cursor.execute(insert_query, (row['space_id'], row['space_nome'], row['space_privado'], row['space_arquivado'], row['status_id'], row['status_descricao'], row['status_tipo'], row['status_index'], row['data_hora']))

# Confirmando as alterações
connection.commit()

# Fechando o cursor atual
cursor.close()

# Fechando conexão
if connection:
            cursor.close()
            connection.close()